@@ -110,6 +110,25 @@ class TestDatastore(unittest.TestCase):
     checkLength(0)
 
 
+class TestNullDatastore(unittest.TestCase):
+
+  def test_null(self):
+
+    s = datastore.NullDatastore()
+
+    for c in range(1, 20):
+      c = str(c)
+      k = Key(c)
+      self.assertFalse(s.contains(k))
+      self.assertEqual(s.get(k), None)
+      s.put(k, c)
+      self.assertFalse(s.contains(k))
+      self.assertEqual(s.get(k), None)
+
+    for item in s.query(Query(Key('/'))):
+      raise Exception('Should not have found anything.')
+
+
 class TestDictionaryDatastore(TestDatastore):
 
   def test_dictionary(self):
@@ -122,6 +141,31 @@ class TestDictionaryDatastore(TestDatastore):
     self.subtest_simple(stores)
 
 
+
+class TestCacheShimDatastore(TestDatastore):
+
+  def test_simple(self):
+    from datastore import CacheShimDatastore
+    from datastore import NullDatastore
+
+    class NullMinusQueryDatastore(NullDatastore):
+      def query(self, query):
+        raise NotImplementedError
+
+    # make sure the cache is used
+    s1 = CacheShimDatastore(NullMinusQueryDatastore(), cache=DictDatastore())
+
+    # make sure the cache is not relief upon
+    s2 = CacheShimDatastore(DictDatastore(), cache=NullDatastore())
+
+    # make sure the cache works in tandem
+    s3 = CacheShimDatastore(DictDatastore(), cache=DictDatastore())
+
+    self.subtest_simple([s1, s2, s3])
+
+
+
+
 class TestKeyTransformDatastore(TestDatastore):
 
   def test_simple(self):
@@ -321,15 +365,15 @@ class TestNestedPathDatastore(TestDatastore):
     test(1, 4, 'abcd')
     test(3, 10, 'abcdefghij/k')
 
-  def test_nested_path_ds(self):
+  def subtest_nested_path_ds(self, **kwargs):
 
-    k1 = Key('/abcdefghijk')
-    k2 = Key('/abcdefghijki')
-    k3 = Key('/ab/cd/ef/abcdefghijk')
-    k4 = Key('/ab/cd/ef/abcdefghijki')
+    k1 = kwargs.pop('k1')
+    k2 = kwargs.pop('k2')
+    k3 = kwargs.pop('k3')
+    k4 = kwargs.pop('k4')
 
     ds = DictDatastore()
-    np = datastore.NestedPathDatastore(ds, depth=3, length=2)
+    np = datastore.NestedPathDatastore(ds, **kwargs)
 
     self.assertFalse(ds.contains(k1))
     self.assertFalse(ds.contains(k2))
@@ -344,8 +388,6 @@ class TestNestedPathDatastore(TestDatastore):
     np.put(k1, k1)
     np.put(k2, k2)
 
-    print ds._items
-
     self.assertFalse(ds.contains(k1))
     self.assertFalse(ds.contains(k2))
     self.assertTrue(ds.contains(k3))
@@ -406,6 +448,44 @@ class TestNestedPathDatastore(TestDatastore):
     self.assertFalse(np.contains(k4))
 
 
+  def test_3_2(self):
+
+    opts = {}
+    opts['k1'] = Key('/abcdefghijk')
+    opts['k2'] = Key('/abcdefghijki')
+    opts['k3'] = Key('/ab/cd/ef/abcdefghijk')
+    opts['k4'] = Key('/ab/cd/ef/abcdefghijki')
+    opts['depth'] = 3
+    opts['length'] = 2
+
+    self.subtest_nested_path_ds(**opts)
+
+  def test_5_3(self):
+
+    opts = {}
+    opts['k1'] = Key('/abcdefghijk')
+    opts['k2'] = Key('/abcdefghijki')
+    opts['k3'] = Key('/abc/def/ghi/jka/bcd/abcdefghijk')
+    opts['k4'] = Key('/abc/def/ghi/jki/abc/abcdefghijki')
+    opts['depth'] = 5
+    opts['length'] = 3
+
+    self.subtest_nested_path_ds(**opts)
+
+  def test_keyfn(self):
+
+    opts = {}
+    opts['k1'] = Key('/abcdefghijk')
+    opts['k2'] = Key('/abcdefghijki')
+    opts['k3'] = Key('/kj/ih/gf/abcdefghijk')
+    opts['k4'] = Key('/ik/ji/hg/abcdefghijki')
+    opts['depth'] = 3
+    opts['length'] = 2
+    opts['keyfn'] = lambda key: key.name[::-1]
+
+    self.subtest_nested_path_ds(**opts)
+
+
 class TestDatastoreCollection(TestDatastore):
 
   def test_tiered(self):
